from .classes import Conversation, Timeline


def make_story(settings):
    """
    Creates web page and data from a twitter account's stream.

    Function extracts needed settings. Then, after, btaining twitter api instance,
    a ``Timeline`` object is instantiated. It's data is encoded into a JSON file. This
    allows portability for the timeline instance.

    The timeline's JSON is then consumed for the creation of a ``Conversation``
    instance.

    The ``Conversation`` instance is passed along with a template file path location
    and a file path for output to a ``write`` function.  The ``write`` function
    takes care of usering conversation data to produce the HTML page that represents
    the "story".

    Finally, if an email handler was included in the settings, then that email
    handler is called; it is passed the location of the just-produced HTML "story"
    page, but it may choose to not use it/attach it.

    Args:
        settings (dict): Configuration settings.

    Returns:
        str: The file path location of the generated web page.
    """
    print('Starting conversationalist. Getting tweets...')
    adapter = settings.get('adapter')
    api = settings['api']
    timeline_json_output_file = settings['timeline_out']
    timeframe_hours = int(settings.get('timeframe', 24))
    title = settings.get('title', 'Story')
    twitter_username = settings['username']
    write = settings['write']
    timeline = Timeline(api, twitter_username, (timeframe_hours * -1))
    print("...saving Timeline as JSON file...")
    timeline.to_json(timeline_json_output_file)
    conversation = Conversation(title=title, adapter=adapter)
    conversation.load(timeline_json_output_file)
    print("...writing story file...")
    page_location = write(conversation, settings['story_out'])
    print('...conversationalist done.')
    return page_location


def print_rate_limit_info(api):
    """
    Prints rate limit information for passed API instance.

    Args:
        api: A tweepy API instance.
    """
    info = api.rate_limit_status()
    for (key, val) in info['resources'].items():
        if key == 'statuses':
            for (status_key, status_value) in val.items():
                print ('KEY: ', status_key, ' VALUE: ', status_value)


from datetime import datetime, timedelta, timezone
import json
import pytz
from operator import attrgetter
from dateutil.parser import parse
from tweepy.error import TweepError
from tweepy.models import User, Status


class UserEncoder(json.JSONEncoder):
    """
    Encodes ``tweepy`` ``User`` objects into an abbreviated JSON format object.

    Encodes the fields ``id``, ``screen_name``, and ``profile_image_url``. And that's
    it.
    """
    def default(self, o):
        user = {
            'id': o.id,
            'screen_name': o.screen_name,
            'profile_image_url': o.profile_image_url
        }
        return user


class StatusEncoder(json.JSONEncoder):
    """
    Encodes `~.classes.Status` objects into JSON.

    The ``created_at`` property is encoded as a string in ISO8601 format.
    """
    def default(self, obj):
        if isinstance(obj, User):
            return UserEncoder().default(obj)
        simple_origin = None
        origin = getattr(obj, 'origin', None)
        if origin:
            simple_origin = {
                'author': UserEncoder().default(obj.origin.author),
                'text': obj.origin.text
            }
        status = {
            'author': obj.author,
            'origin': simple_origin,
            'text': obj.text,
            'created_at': obj.created_at.isoformat(),
            'in_reply_to_status_id': getattr(obj, 'in_reply_to_status_id', '')
        }
        return status


class TimelineEncoder(json.JSONEncoder):
    """
    Encodes `~.classes.Timeline` objects into JSON.

    The ``start`` and ``cutoff`` properties are encoded as strings in ISO8601 format.
    """
    def default(self, obj):
        if isinstance(obj, User):
            return UserEncoder().default(obj)
        if isinstance(obj, Status):
            return StatusEncoder().default(obj)
        timeline = {
            'start': obj.start.isoformat(),
            'cutoff': obj.cutoff.isoformat(),
            'data': obj.data,
            'total': obj.total,
            'username': obj.username
        }
        return timeline


class Participant(object):
    def __init__(self, name, profile_url=None):
        self.exchange_count = 0
        self.name = name
        self.profile_url = profile_url

    def increment_participation(self):
        self.exchange_count += 1


class Participation(object):
    def __init__(self):
        self.participants = {}

    def add_tweet(self, author):
        if author['screen_name'] in self.participants:
            participant = self.participants[author['screen_name']]
            participant.increment_participation()
        else:
            participant = Participant(author['screen_name'], author['profile_image_url'])
            participant.increment_participation()
            self.participants[author['screen_name']] = participant

    def get_ranked_profiles(self):
        ranked_profiles = []
        for key in self.participants:
            ranked_profiles.append(self.participants[key])
        ranked_profiles = sorted(ranked_profiles, key=attrgetter('exchange_count'), reverse=True)
        return ranked_profiles


class Conversation(object):
    """
    Manages the state of user's tweet stream during application processing.
    """
    def __init__(self, timeline=None, title='Tick Tock', adapter=None):
        """
        Initializes a ``Conversation`` object.

        When ``timeline``, ``start``, and ``cutoff`` data is passed, the
        class internally undertakes the exchange of status data into
        the conversation's format.  Otherwise, it just holds the
        state of whatever property data is passed.

        Args:
            timeline (dict): A JSON object representing a timeline.
            title (str): The name for the conversation object's data.
            adapter: Handles transformation logic for status data.

        Attributes:
            title (str): Main name for conversation.
            adapter: Handles transformation logic for status data.
            data (dict): Contains pairing of timestamps and statuses keyed to 'hourlies', as
                well as a title name keyed to 'title'.
        """
        self.data = None
        self.timeline = timeline
        self.title = title
        self.adapter = adapter
        self.update_conversation()

    def _get_timeline_interval(self):
        if self.timeline:
            start = parse(self.timeline['start'])
            cutoff = parse(self.timeline['cutoff'])
            return start, cutoff
        return None, None

    def update_conversation(self):
        """
        Applies the class instances adapter to the current timeline data.
        """
        if self.timeline and self.adapter:
            adapter = self.adapter(self)
            self.data = adapter.convert()

    def load(self, json_file):
        """
        Transforms the JSON data for a user timeline into
        relevant properties for this class.

        Args:
            json_file (str): The file location of the timeline JSON.
        """
        with open(json_file) as infile:
            timeline_json = json.load(infile)
        self.timeline = timeline_json
        self.update_conversation()


class Timeline(object):
    """
    Manages state of a a twitter user's timeline data.

    Attributes:
        api: Tweepy API instance.
        earliest_status: Holds the earliest status handled during timeline generation.
        start (datetime): When the timeline starts. Set to ``now`` at initialization.
        cutoff (datetime): When in the past the timeline's search for statuses ends.
        data (dict): Maps an identifier to status information.
        username (str): The targeted account's username.
    """
    def __init__(self, api=None, username=None, timeframe=-24):
        self.api = api
        self.earliest_status = None
        self.start = datetime.now(tz=timezone.utc)
        safe_timeframe = abs(timeframe) * -1
        self.cutoff = self.start + timedelta(hours=safe_timeframe)
        self.data = {}
        self.username = username
        if api and username:
            self._generate_timeline()

    def load(self, statuses):
        """
        The method transforms and appends the passed statuses to the class instances
        `statuses` property.

        If a status is a response, the method searches for the original
        tweet an adds the text and author name to the status data.

        Raises:
            TweepError: If tweepy error occurs while seeking an origin for a
                status, which is typically a tweet that was responded too.
        Args:
            statuses (list): A a list of tweepy ``Status`` objects.
        """
        for status in statuses:
            if str(status.id) not in self.data:
                # check if 'created_at' is naive. Tweepy creates naive created_at fields
                # from utc timestamps parsed from Twitter API's RFC 2822 format
                if status.created_at.tzinfo is None or \
                        status.created_at.tzinfo.utcoffset(status.created_at) is None:
                    utc = pytz.utc
                    # naive to utc
                    utc_created_at = utc.localize(status.created_at)
                    status.created_at = utc_created_at
                if status.created_at > self.cutoff:
                    #status.text = status.text.encode('ascii', 'ignore')
                    status.origin = None
                    if status.in_reply_to_status_id:
                        try:
                            status.origin = self.api.get_status(status.in_reply_to_status_id)
                            #status.origin.text = status.origin.text.encode('ascii', 'ignore')
                            user = status.origin.author
                            status.origin.author_name = user.screen_name
                        except TweepError:
                            print('Error while fetching origin for tweet {0}'.format(status.id))
                    self.data[str(status.id)] = status

    def get_earliest_status(self):
        """
        Sorts the class instance's `statuses` by their `created_at`
        property from oldest at the start of the list to newest
        at the end of the list.

        Returns:
            ``Status``: The earliest (oldest) status or ``None`` if empty.
        """
        earliest = None
        if self.data:
            sorted_statuses = sorted(self.data.values(), key=lambda status: status.created_at)
            earliest = sorted_statuses[0]
        return earliest

    @property
    def total(self):
        """
        Computed property representing the total count of statuses for the instance.

        Returns:
            int: The total count of statuses.
        """
        return len(list(self.data.values()))

    def get_timeline_batch(self, max_id=None):
        """
        Gets a batch of statuses for a user.

        If a ``max_id`` is passed, that is supplied to the API
        is a starting point for fetching the previous 20 tweets.

        Args:
            max_id: The last identifier that included in this batch
              of statuses.

        Returns:
            list: A list of tweepy ``Status`` objects. The maximum size of
            the list is 20.
        """
        if max_id is None:
            return self.api.user_timeline(self.username)
        else:
            return self.api.user_timeline(self.username, max_id=max_id)

    def _has_next_tweets(self):
        """
        Checks current statuses to see if the timeframe cutoff for a user's timeline
        has been reached or available statuses have been exhausted.

        Returns:
            bool: ``True`` if additional twitter statuses that fit into the timeframe cutoff
                may exist. ``False`` otherwise.
        """
        new_earliest_status = self.get_earliest_status()
        if new_earliest_status:
            # Exceeded cutoff because the earliest status is older (lesser) than the cutoff
            if new_earliest_status.created_at < self.cutoff:
                self.earliest_status = new_earliest_status
                return False
            # In case all available tweets are exhausted
            if self.earliest_status and \
                    self.earliest_status.id == new_earliest_status.id:
                return False
            # keep going - cutoff not exceeded and new earliest id not the current earliest id
            self.earliest_status = new_earliest_status
            return True
        return False

    def _generate_timeline(self):
        """
        Called during instance intialization. It fetches tweet statuses allowed
        by the cutoff timeframe until available statuses are exhausted.
        """
        tweets_available = True
        while tweets_available:
            earliest_id = getattr(self.earliest_status, 'id', None)
            new_tweets = self.get_timeline_batch(earliest_id)
            if new_tweets:
                self.load(new_tweets)
                tweets_available = self._has_next_tweets()
            else:
                tweets_available = False

    def to_json(self, file_path):
        """
        Writes a JSON file base on instance data.

        Args:
            file_path (str): Where the JSON file will be written.
        """
        with open(file_path, 'w') as outfile:
            json.dump(self, outfile, cls=TimelineEncoder, indent=2)
        return file_path



import re
from datetime import datetime, timedelta
from operator import itemgetter
from dateutil.parser import parse
from .classes import Participation

PERIOD_DT_FORMAT = '%A, %B %d, %Y  %-I%p'


def initialize_hourly_summary(start, cutoff):
    """
    Generates a dict that contains statuses for each hour during
    a timeframe. The statuses are keyed to a timestamp string.

    At initialization, the list for each timestamp is empty.

    Args:
        start (datetime): The timeline's start.
        cutoff (datetime): When the timeline's status search ends.

    Returns:
        dict: Statuses keyed to timestamps arranged in hourly increments.
    """
    hourly_summary = {}
    start = start.replace(minute=0)
    active = cutoff
    while active < start:
        hourly_key = active
        hourly_key = hourly_key.replace(minute=0, second=0, microsecond=0)
        hourly_summary[hourly_key.isoformat()] = []
        active = active + timedelta(hours=1)
    return hourly_summary


def sort_statuses(statuses):
    return sorted(statuses, key=lambda s: s['created_at'])


def to_periods(hourly_summary):
    periods = []
    for iso_timestamp, statuses in hourly_summary.items():
        period_datetime = parse(iso_timestamp)
        unix_epoch = datetime(1970, 1, 1, tzinfo=period_datetime.tzinfo)
        seconds = (period_datetime - unix_epoch).total_seconds()
        subtitle = period_datetime.strftime(PERIOD_DT_FORMAT)
        if len(statuses) > 0:
            empty = False
            message = 'No updates.'
            hour_block = {
                'id': int(seconds),
                'empty': empty,
                'empty_message': message,
                'subtitle': subtitle,
                'statuses': sort_statuses(statuses)
            }
            periods.append(hour_block)
    periods = sorted(periods, key=itemgetter('id'))
    return periods


def find_topic_header(status, pattern, return_group=0):
    """
    Searches a status for the presence of a pattern, and returns matches


    Args:
        status(dict): A dictionary with information for a tweet. Expected to
          include a ``text`` key.
        pattern: A raw string with a regular expression pattern.
        return_group (int): The regex group that may be returned. Defaults to all groups.
    Returns:
        The requested return group for the match or ``None``.
    """
    text = status['text']
    topic_regex = re.compile(pattern, re.I)
    match = topic_regex.search(text)
    if match:
        return match.group(return_group)
    return None


def get_style_classes(style_words, status):
    """
    Searches for occurrence of key words in status text; the matches
    will be used by templates to apply those word matches as CSS styles.

    Args:
        style_words (list): The targeted search words.
        status (dict): Data for a tweet status.

    Returns:
        str: The matched classes or an empty string if no matches found.
    """
    style_classes = ''
    style_matches = []
    if style_words:
        for word in style_words:
            pattern = r'\b%s\b' % word
            regex = re.compile(pattern, re.I)
            match = regex.search(status['text'])
            if match:
                word = word.replace(' ', '-')
                style_matches.append(word)
    for m in set(style_matches):
        style_classes = '{0} {1}'.format(style_classes, m)
    return style_classes.strip()


def transform_with_topic_headers(conversation, pattern, return_goup):
    start, cutoff = conversation._get_timeline_interval()
    hourly_summary = initialize_hourly_summary(start, cutoff)
    timeline_data = conversation.timeline.get('data', {})
    topic_headers = []
    for identifier, status in timeline_data.items():
        if pattern:
            topic_header = find_topic_header(status, pattern, return_goup)
            if topic_header:
                status['topic_header'] = topic_header
                topic_headers.append(topic_header)
        created_with_no_minutes = parse(status['created_at']).replace(minute=0, second=0, microsecond=0)
        time_key = created_with_no_minutes.isoformat()
        if time_key in hourly_summary:
            # statuses not sorted by time
            hourly_summary[time_key].append(status)
        else:
            hourly_summary[time_key] = [status]
    nav = sorted(set(topic_headers))
    data = {
        'title': conversation.title,
        'periods': to_periods(hourly_summary),
        'topic_headers': nav
    }
    return data


def transform_with_participation_and_styles(conversation, style_words,
                                            header_pattern, return_group):
    """
    Iterates through conversation status dictionaries adding their data to
    the instances ``Participation`` property, handling content transformations,
    and inserting style classes.  This logic helps create more informative
    pages once the data is rendered.
    """
    start, cutoff = conversation._get_timeline_interval()
    hourly_summary = initialize_hourly_summary(start, cutoff)
    timeline_data = conversation.timeline.get('data', {})
    participation = Participation()
    topic_headers = []
    for identifier, status in timeline_data.items():
        participation.add_tweet(status['author'])
        if status['origin']:
            participation.add_tweet(status['origin']['author'])
        if header_pattern:
            topic_header = find_topic_header(status, header_pattern, return_group)
            if topic_header:
                status['topic_header'] = topic_header
                topic_headers.append(topic_header)
        if style_words:
            status['style_classes'] = get_style_classes(style_words, status)
        created_with_no_minutes = parse(status['created_at']).replace(minute=0, second=0, microsecond=0)
        time_key = created_with_no_minutes.isoformat()
        if time_key in hourly_summary:
            # statuses not sorted by time
            hourly_summary[time_key].append(status)
        else:
            hourly_summary[time_key] = [status]
    nav = sorted(set(topic_headers))
    data = {
        'title': conversation.title,
        'periods': to_periods(hourly_summary),
        'participation': participation,
        'nav': nav
    }
    return data


class ParticipationAdapter:
    style_words = None
    header_pattern = None
    return_group = 0

    def __init__(self, conversation):
        self.conversation = conversation

    def convert(self):
        return transform_with_participation_and_styles(self.conversation,
                                                       self.style_words,
                                                       self.header_pattern,
                                                       self.return_group)


class TopicHeaderAdapter:
    pattern = None
    return_group = 0

    def __init__(self, conversation):
        self.conversation = conversation

    def convert(self):
        return transform_with_topic_headers(self.conversation, self.pattern, self.return_group)


class TextReplaceAdapter:
    conversions = None

    def __init__(self, conversation):
        self.conversation = conversation

    def convert(self):
        start, cutoff = self.conversation._get_timeline_interval()
        hourly_summary = initialize_hourly_summary(start, cutoff)
        timeline_data = self.conversation.timeline.get('data', {})
        for identifier, status in timeline_data.items():
            if self.conversions:
                for original, replacement in self.conversions.items():
                    status['text'].replace(original, replacement)
            created_with_no_minutes = parse(status['created_at']).replace(minute=0, second=0, microsecond=0)
            time_key = created_with_no_minutes.isoformat()
            if time_key in hourly_summary:
                # statuses not sorted by time
                hourly_summary[time_key].append(status)
            else:
                hourly_summary[time_key] = [status]
        data = {
            'title': self.conversation.title,
            'periods': to_periods(hourly_summary),
        }
        return data



#!/usr/bin/env python3
import os
from setuptools import setup, find_packages


def get_readme():
    return open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

setup(
    author="Julio Gonzalez Altamirano",
    author_email='devjga@gmail.com',
    description="Converts tweet streams into an organized, story-telling web-page .",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3 :: Only'
    ],
    install_requires=['pytz', 'python-dateutil', 'tweepy'],
    keywords="python twitter",
    license="MIT",
    long_description=get_readme(),
    name='conversationalist',
    packages=find_packages(include=['conversationalist', 'conversationalist.*'],
                           exclude=['tests', 'tests.*']),
    platforms=['Any'],
    url='https://github.com/jga/conversationalist',
    version='0.1.1',
)